/*****************************************************************************
 * Local Headers
 *****************************************************************************/

#include <windows.h>

#include "connectlog.h"
#include "util.h"
#include "dir.h"

#define FW (16)

/*****************************************************************************
 * Class: ConnectLog
 *****************************************************************************/

ConnectLog::ConnectLog(
   const string &dir,
   const string &ip,
   int   threadID,
   int   connectNumber
) :
   mDir(dir),
   mIP(ip),
   mThreadID(threadID),
   mConnectNumber(connectNumber),

   mFullPath(""),
   mConnectLogName(""),
   mWroteConnectLog(false),

   mTotalRequestBytes(0),
   mTotalResponseBytes(0),
   mTotalStartSeconds(0),
   mTotalStartMillis(0),

   mSessionNumber(0),
   mIsConnect(false),
   mHost(""),
   mPort(0),

   mRequestBytes(0),
   mResponseBytes(0),
   mStartSeconds(0),
   mStartMillis(0),
   mSessionActive(false),

   mRequestSeekNN(false),
   mResponseSeekNN(false),
   mRequestExt(""),
   mResponseExt("")
{
   Util::now(&mTotalStartSeconds, &mTotalStartMillis);
}

ConnectLog::~ConnectLog()
{
   if(mWroteConnectLog)
   {
      FILE *f = NULL;

      if((f = fopen(mConnectLogName.c_str(), "a")) == NULL)
      {
         // Marginally better than nothing ...

         fprintf(
            stderr,
            "(%s|%d) Could not open file: [%s]\n",
            __FILE__, __LINE__,
            mConnectLogName.c_str()
         );
      }
      else
      {
         int endSeconds = 0;
         int endMillis  = 0;

         Util::now(&endSeconds, &endMillis);

         string end          = Util::getTimeStamp(endSeconds);
         double endTimeStamp = endSeconds + endMillis / 1000.0;

         fprintf(f, "%-*s = %s\n",   FW, "end",            end.c_str());
         fprintf(f, "%-*s = %.3f\n", FW, "end_timestamp",  endTimeStamp);
         fprintf(f, "%-*s = %d\n",   FW, "request_bytes",  mTotalRequestBytes);
         fprintf(f, "%-*s = %d\n",   FW, "response_bytes", mTotalResponseBytes);

         fclose(f);
      }
   }

   if(mSessionActive)
   {
      string errMsg;

      if(!endSession(errMsg))
      {
         fprintf(stderr, "(%s|%d) %s\n", __FILE__, __LINE__, errMsg.c_str());
      }
   }
}

bool ConnectLog::beginSession(
   bool isConnect, const string &host, int port, string &errMsg
)
{
   if(!mWroteConnectLog)
   {
      mFullPath =
         mDir + Dir::csep() + Util::toString("%05d", mConnectNumber);

      if(!Dir::ensureDir(mFullPath, errMsg))
         return(false);

      mConnectLogName = mFullPath + Dir::csep() + "connect.log";
      FILE *f         = NULL;

      if((f = fopen(mConnectLogName.c_str(), "w")) == NULL)
      {
         errMsg = "Could not open file: [";
         errMsg += mConnectLogName + "]";

         return(false);
      }

      string start          = Util::getTimeStamp(mTotalStartSeconds);
      double startTimeStamp = mTotalStartSeconds + mTotalStartMillis / 1000.0;

      fprintf(f, "%-*s = %d\n",   FW, "connect_number",  mConnectNumber);
      fprintf(f, "%-*s = %d\n",   FW, "thread_id",       mThreadID);
      fprintf(f, "%-*s = %s\n",   FW, "ip_address",      mIP.c_str());
      fprintf(f, "%-*s = %s\n",   FW, "start",           start.c_str());
      fprintf(f, "%-*s = %.3f\n", FW, "start_timestamp", startTimeStamp);

      fclose(f);

      mWroteConnectLog = true;
   }

   if(mSessionActive)
   {
      if(!endSession(errMsg))
         return(false);
   }

   ++mSessionNumber;

   mIsConnect     = isConnect;
   mHost          = host;
   mPort          = port;
   mRequestBytes  = 0;
   mResponseBytes = 0;
   mRequestSeekNN = true;
   mRequestExt    = "log";

   if(mIsConnect)
   {
      mResponseSeekNN = false;
      mResponseExt    = "dat";
   }
   else
   {
      mResponseSeekNN = true;
      mResponseExt    = "log";
   }

   Util::now(&mStartSeconds, &mStartMillis);

   mSessionActive = true;

   return(true);
}

static bool findNN(const char *buffer, int n, int *nn)
{
   for(int i = 0; i < n; i ++)
   {
      if(buffer[i] == '\r' && i + 3 < n)
      {
         if(buffer[i + 1] == '\n' &&
            buffer[i + 2] == '\r' &&
            buffer[i + 3] == '\n')
         {
            *nn = i + 4;
            return(true);
         }
      }
   }

   return(false);
}

#if 0
# define Z do { \
   printf("[%s|%d] [%d]\n", __FILE__, __LINE__, GetCurrentThreadId()); \
  } while(0)
#else
# define Z do {} while(0)
#endif

bool ConnectLog::requestBytes(
   const char *buffer, int n, string &errMsg
)
{
Z;
   string base = mFullPath + Dir::csep() +
      Util::toString("req_%03d", mSessionNumber) + ".";

   string fileName = base + mRequestExt;
   FILE *f         = NULL;

   if((f = fopen(fileName.c_str(), "ab")) == NULL)
   {
      errMsg = "Could not open file: [";
      errMsg += fileName + "] for appending";

      return(false);
   }
Z;
   int bytesToWrite = n;

   if(mRequestSeekNN)
   {
      int nn = 0;

      if(findNN(buffer, n, &nn))
      {
         mRequestExt    = "dat";
         mRequestSeekNN = false;
         bytesToWrite   = nn;
      }
   }
Z;
   if(!fwrite(buffer, bytesToWrite, 1, f))
   {
      fclose(f);
      errMsg = "Error writing to file: [";
      errMsg += fileName + "]";
      return(false);
   }
Z;
   fclose(f);

   // Sometimes we'll find our 'NN' boundary but the remainder
   // of 'n' begins with a null (\0) character.  In those cases
   // we don't output the bit after the 'NN' boundary.

   int addBytes = n;

   if(bytesToWrite < n)
   {
      if(buffer[bytesToWrite])
      {
         fileName    = base + mRequestExt;
         f           = NULL;

         if((f = fopen(fileName.c_str(), "ab")) == NULL)
         {
            errMsg = "Could not open file: [";
            errMsg += fileName + "] for appending";

            return(false);
         }
Z;
         if(!fwrite(buffer + bytesToWrite, n - bytesToWrite, 1, f))
         {
            fclose(f);
            errMsg = "Error writing to file: [";
            errMsg += fileName + "]";
            return(false);
         }

         fclose(f);
      }
      else
      {
         addBytes = bytesToWrite;
      }
   }
Z;
   mTotalRequestBytes += addBytes;
   mRequestBytes      += addBytes;
Z;
   return(true);
}

bool ConnectLog::responseBytes(
   const char *buffer, int n, string &errMsg
)
{
Z;
   string base = mFullPath + Dir::csep() +
      Util::toString("res_%03d", mSessionNumber) + ".";

   string fileName = base + mResponseExt;
   FILE *f         = NULL;
Z;
   if((f = fopen(fileName.c_str(), "ab")) == NULL)
   {
      errMsg = "Could not open file: [";
      errMsg += fileName + "] for appending";

      return(false);
   }
Z;
   int bytesToWrite = n;

   if(mResponseSeekNN)
   {
      int nn = 0;

      if(findNN(buffer, n, &nn))
      {
         mResponseSeekNN = false;
         bytesToWrite   = nn;
      }
   }
Z;
   if(!fwrite(buffer, bytesToWrite, 1, f))
   {
      fclose(f);
      errMsg = "Error writing to file: [";
      errMsg += fileName + "]";
      return(false);
   }
Z;
   fclose(f);

   if(bytesToWrite < n)
   {
      mResponseExt = "dat";
      fileName    = base + mResponseExt;
      f           = NULL;

      if((f = fopen(fileName.c_str(), "ab")) == NULL)
      {
         errMsg = "Could not open file: [";
         errMsg += fileName + "] for appending";

         return(false);
      }
Z;
      if(!fwrite(buffer + bytesToWrite, n - bytesToWrite, 1, f))
      {
         fclose(f);
         errMsg = "Error writing to file: [";
         errMsg += fileName + "]";
         return(false);
      }

      fclose(f);
   }
Z;
   mTotalResponseBytes += n;
   mResponseBytes      += n;
Z;
   return(true);
}

bool ConnectLog::endSession(string &errMsg)
{
   int endSeconds = 0;
   int endMillis  = 0;

   Util::now(&endSeconds, &endMillis);

   double startTimeStamp = mStartSeconds + mStartMillis / 1000.0;
   double endTimeStamp   = endSeconds    + endMillis    / 1000.0;

   string fileName =
      mFullPath + Dir::csep() +
      Util::toString("inf_%03d.log", mSessionNumber);

   FILE *f = NULL;

   if((f = fopen(fileName.c_str(), "w")) == NULL)
   {
      errMsg = "Could not open: [";
      errMsg += fileName + "]";

      return(false);
   }

   fprintf(f, "host            = %s\n", mHost.c_str());
   fprintf(f, "port            = %d\n", mPort);
   fprintf(f, "connect         = %s\n", mIsConnect ? "yes" : "no");
   fprintf(f, "start_timestamp = %f\n", startTimeStamp);
   fprintf(f, "end_timestamp   = %f\n", endTimeStamp);
   fprintf(f, "request_bytes   = %d\n", mRequestBytes);
   fprintf(f, "response_bytes  = %d\n", mResponseBytes);

   fclose(f);

   return(true);
}
