/*****************************************************************************
 * Headers
 *****************************************************************************/

#include <stdio.h>
#include <sstream>

using std::ostringstream;

#include "util.h"
#include "output.h"
#include "mutexmanager.h"

/*****************************************************************************
 * Class Definition: Output
 *****************************************************************************/

Output::Output(bool doVerbose) : mVerbose(doVerbose)
{
}

Output::Output() : mVerbose(false)
{
}

Output::~Output()
{
}

void Output::exit(const string &s)
{
   write(s);
   ::exit(-1);
}

void Output::write(const string &s)
{
}

void Output::writeVerbose(const string &s)
{
   if(mVerbose)
      write(s);
}

string Output::info(const char *file, int line)
{
   return(Output::info(file, line, ""));
}

string Output::info(const char *file, int line, string msg)
{
   return(Output::info(file, line, 0, msg));
}

string Output::info(const char *file, int line, int id, string msg)
{
   string now   = Util::nowString();
   int dotIndex = Util::indexOf(now, '.');
   int nDigits  = 4;

   while(dotIndex - 1 >= 0)
   {
      dotIndex--;

      if(--nDigits == 0)
         break;
   }

   ostringstream oss;

   oss
      << "("
      << Util::substring(now, dotIndex)
      << "|"
      << file
      << "|"
      << line;

   if(id > 0)
      oss << "|" << id;

   oss << ")";

   if(msg != "")
      oss << " " << msg;

   return(oss.str());
}

// *************************
// MutexOutput
// *************************

MutexOutput::MutexOutput(HANDLE h, bool doVerbose) :
   Output(doVerbose), mH(h)
{
}

void MutexOutput::write(const string &s)
{
   string errMsg = "";

   //printf("JFD: %d => acquiring lock ...\n", GetCurrentThreadId());

   if(!MutexManager::acquire(mH, errMsg))
   {
      fprintf(stderr, "Could not acquire mutex for output\n");
   }
   else
   {
      //printf("JFD: %d => acquired lock ...\n", GetCurrentThreadId());

      writeSynchronized(s);
      //Sleep(2 * 1000);

      if(!MutexManager::release(mH, errMsg))
      {
         fprintf(stderr, "Could not release mutex for output\n");
      }

      //printf("JFD: %d => released lock ...\n", GetCurrentThreadId());
   }
}

// ---------------------------
// Local helper function ...
// ---------------------------

static void doPrint(FILE *f, const string &s)
{
   fprintf(f, "%s\n", s.c_str());
}

// *************************
// ScreenOutput
// *************************

ScreenOutput::ScreenOutput(HANDLE mutex, bool doVerbose) :
   MutexOutput(mutex, doVerbose)
{
}

void ScreenOutput::writeSynchronized(const string &s)
{
   doPrint(stdout, s);
}

// *************************
// LogOutput
// *************************

LogOutput::LogOutput(
   const string &fileName, HANDLE mutex, bool doVerbose
) : MutexOutput(mutex, doVerbose), mName(fileName)
{
}

void LogOutput::writeSynchronized(const string &s)
{
   FILE *f = NULL;

   if((f = fopen(mName.c_str(), "a")) == NULL)
   {
      fprintf(
         stderr,
         "Could not open: [%s], Msg: [%s]\n",
         mName.c_str(), s.c_str()
      );
   }
   else
   {
      doPrint(f, s);
      fclose(f);
   }
}

// *************************
// TeeOutput
// *************************

TeeOutput::TeeOutput(Output *a, Output *b) : mA(a), mB(b)
{
}

TeeOutput::~TeeOutput()
{
   delete mA;
   delete mB;
}

void TeeOutput::writeVerbose(const string &s)
{
   mA->writeVerbose(s);
   mB->writeVerbose(s);
}

void TeeOutput::write(const string &s)
{
   mA->write(s);
   mB->write(s);
}
