/*****************************************************************************
 * Headers / Using Declarations
 *****************************************************************************/

#include <winsock2.h>
#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>

#include <sstream>
#include <string>
#include <vector>
#include <memory>

#include "clientthreadhandler.h"
#include "handlecleanup.h"
#include "mutexmanager.h"
#include "threaddata.h"
#include "connectlog.h"
#include "output.h"
#include "logger.h"
#include "config.h"
#include "util.h"
#include "cfg.h"
#include "dir.h"

using std::ostringstream;
using std::auto_ptr;
using std::vector;
using std::string;

// Request that the library we need (ws2_32.lib) be
// automatically included at link time.

#pragma comment(lib, "ws2_32.lib")

/*****************************************************************************
 * Macros
 *****************************************************************************/

#define _PC(obj) obj->getPrefix(__LINE__)

/*****************************************************************************
 * Local Function Prototypes
 *****************************************************************************/

static bool goToWorkingDirectory(string &errMsg);

static bool getConfig(
   const string &fileName, Config &config, string &errMsg
);

static Output *getOutput(
   MutexManager &mgr,
   const string &outputDir,
   bool         doVerbose,
   bool         doScreen,
   bool         doLog
);

static bool getOutputDirectory(
   string& outputDir, const string &dir, string &errMsg
);

static bool   doStartup (string &errMsg);
static void   setTimeOut(double timeOutSeconds, struct timeval *ptr);
static string getBanner (const string &port);
static void   bail      (const string &errMsg);

static bool processRequest(
   const char *buffer,
   string     &type,
   bool       &isConnect,
   string     &host,
   int        &port,
   Output     *o,
   int        tID,
   string     &errMsg
);

static int clientThread(ThreadData *threadData);

/*****************************************************************************
 * Local Functions
 *****************************************************************************/

static bool goToWorkingDirectory(string &errMsg)
{
   // We always want to run in the directory where
   // our exe resides so work out the path to
   // ourself and cd there.

   string programPath = "";

   if(!Dir::getProgramPath(programPath, errMsg))
      return(false);

   if(false)
   {
      printf("Program Path: [%s]\n", programPath.c_str());
   }

   string programDir = Util::substring(
      programPath,
      0,
      1 + Util::lastIndexOf(programPath, Dir::csep())
   );

   programDir = Dir::toStdForm(programDir);

   if(false)
   {
      printf("Program Dir: [%s]\n", programDir.c_str());
   }

   if(!Dir::cd(programDir, errMsg))
      return(false);

   return(true);
}

static bool getConfig(
   const string &fileName, Config &config, string &errMsg
)
{
   string defaultPortString    = Util::toString(Config::DEFAULT_PORT);
   string defaultTimeoutString = Util::toString(Config::DEFAULT_TIMEOUT);

   CfgItemSpec spec[] =
   {
      {
         CONFIG_SETTINGS,
         NULL,
         CFG_EXACTLY_ONE,
         NULL,
         Config::settingsBegin,
      },
      {
         CONFIG_SETTINGS,
         CONFIG_PORT,
         CFG_ZERO_OR_ONE,
         defaultPortString.c_str(),
         Config::settingsPort
      },
      {
         CONFIG_SETTINGS,
         CONFIG_DIR,
         CFG_ZERO_OR_ONE,
         NULL,
         Config::settingsDir
      },
      {
         CONFIG_SETTINGS,
         CONFIG_RLOG,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::settingsRLog
      },
      {
         CONFIG_SETTINGS,
         CONFIG_VERBOSE,
         CFG_ZERO_OR_ONE,
         CFG_NO,
         Config::settingsVerbose
      },
      {
         CONFIG_SETTINGS,
         CONFIG_TIMEOUT,
         CFG_ZERO_OR_ONE,
         defaultTimeoutString.c_str(),
         Config::settingsTimeout
      },
      {
         CONFIG_SETTINGS,
         CONFIG_CONSOLE,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::settingsConsole
      },
      {
         CONFIG_SETTINGS,
         CONFIG_LOG,
         CFG_ZERO_OR_ONE,
         CFG_NO,
         Config::settingsLog
      }
   };

   int nSpec = USZ(spec);
   void *v   = (void *)&config;

   if(false)
      config.dump("Initial");

   if(!Cfg::applyDefaults(v, spec, nSpec, errMsg))
      return(false);

   if(false)
      config.dump("After Defaults");

   CfgLines lines;

   if(!Cfg::getLines(fileName.c_str(), lines, errMsg))
      return(false);

   if(false)
      lines.dump();

   CfgSections sections;

   if(!Cfg::parse(lines, sections, errMsg))
      return(false);

   if(false)
   {
      for(int n = sections.size(), i = 0; i < n; i ++)
         sections[i].dump();
   }

   if(!Cfg::processConfig(v, 0, sections, spec, nSpec, errMsg))
      return(false);

   if(false)
      config.dump("Final");

   return(true);
}

static Output *getOutput(
   MutexManager &mgr,
   const string &outputDir,
   bool         doVerbose,
   bool         doScreen,
   bool         doLog
)
{
   Output *out       = NULL;
   Output *outScreen = NULL;
   Output *outLog    = NULL;

   if(doScreen)
   {
      outScreen = new ScreenOutput(mgr.screen(), doVerbose);
   }

   if(doLog)
   {
      string logName =
         outputDir + Dir::csep() + Config::PROGRAM_NAME + ".log";

      outLog = new LogOutput(logName, mgr.log(), doVerbose);
   }

   if(outScreen)
   {
      out = (outLog ? new TeeOutput(outScreen, outLog) : outScreen);
   }
   else if(outLog)
   {
      out = outLog;
   }
   else
   {
      out = new Output(doVerbose);
   }

   return(out);
}

static bool getOutputDirectory(
   string &outputDir, const string &dir, string &errMsg
)
{
   if(dir == "")
   {
      outputDir = Util::getTimeStampDir();
   }
   else
   {
      outputDir = Dir::toStdForm(dir);
   }

   return(Dir::ensureDir(outputDir, errMsg) ? true : false);
}

static bool doStartup(string &errMsg)
{
   WSADATA wsaData;

   int returnValue = WSAStartup(MAKEWORD(2, 0), &wsaData);

   if(returnValue != 0)
   {
      ostringstream oss;

      oss <<
         "Call to WSAStartup failed, returnValue: [" << returnValue << "]";

      errMsg = oss.str();

      return(false);
   }

   return(true);
}

static void setTimeOut(double timeOutSeconds, struct timeval *ptr)
{
   // The 'struct timeval' structure:
   //
   //    tv_sec  => number of full seconds
   //    tv_usec => remainder, millionths of a second
   //
   // So, 4.25 seconds would be represented as:
   //
   //    tv_sec  => 4
   //    tv_usec => 250000

   if(timeOutSeconds <= 0.0)
   {
      // Default to Config::DEFAULT_TIMEOUT for values
      // of timeOutSeconds that are <= 0

      timeOutSeconds = Config::DEFAULT_TIMEOUT;
   }

   int nSec    = (int)timeOutSeconds;
   double rest = timeOutSeconds - nSec;
   int nUSec   = 0;

   if(rest > 0.0)
      nUSec = (int)(1000000.0 * rest + 0.5);

   ptr->tv_sec  = nSec;
   ptr->tv_usec = nUSec;
}

static string getBanner(const string &port)
{
   ostringstream oss;

   oss <<
      Config::PROGRAM_NAME           << ", " <<
      "version: " << Config::VERSION << ", " <<
      "port: "    << port            << ", " <<
      "started: " << Util::getTimeStamp();

   string bar = "";

   for(int i = 0; i < 70; i ++)
      bar += "=";

   return(bar + "\n" + oss.str() + "\n" + bar + "\n");
}

static void bail(const string &errMsg)
{
   printf("%s\n", errMsg.c_str());
   exit(-1);
}

static string toString(ostringstream &oss)
{
   string s = oss.str();
   oss.str("");
   return(s);
}

static bool processRequest(
   const char *buffer,
   string     &type,
   bool       &isConnect,
   string     &host,
   int        &port,
   Output     *o,
   int        tID,
   string     &errMsg
)
{
   // Our first line should contain the request, something like ...
   //
   //    GET http://www.usatoday.com/ HTTP/1.1
   //    CONNECT www.somebank.com:443 HTTP/1.1
   //
   // Determine the host/port we're to connect to
   // as well as whether we're using CONNECT or not

   int index = Util::indexOf(buffer, '\n');

   ostringstream oss;

   if(index == -1)
   {
      oss << "No newline found in: [" << buffer << "]";
      errMsg = oss.str();
      return(false);
   }

   string first = Util::trim(
      Util::substring(buffer, 0, index + 1)
   );

   vector<string> parts = Util::split(first, ' ');

   if(parts.size() != 3)
   {
      oss << "Invalid request: [" << buffer << "]";
      errMsg = oss.str();
      return(false);
   }

   oss <<
      "Parts: [" << parts[0] + "], [" << parts[1] << "], " <<
      "[" << parts[2] << "]";

   o->writeVerbose(_OFLI(tID, toString(oss)));

   type      = parts[0];
   host      = parts[1];
   isConnect = (parts[0] == "CONNECT");

   if(isConnect == false)
   {
      const char *s = "http://";

      if(!Util::startsWith(host, s))
      {
         oss << "Invalid protocol encountered: [" << first << "]";
         errMsg = oss.str();
         return(false);
      }

      host = Util::substring(host, strlen(s));
   }

   if((index = Util::indexOf(host, '/')) != -1)
      host = Util::substring(host, 0, index);

   oss << "Host: [" << host << "]";
   o->writeVerbose(_OFLI(tID, toString(oss)));

   port = 80;

   if((index = Util::indexOf(host, ':')) != -1)
   {
      string theString = "";

      for(int n = host.size(), i = index + 1; i < n; i ++)
      {
         if(host[i] >= '0' && host[i] <= '9')
         {
            theString += host[i];
         }
         else
         {
            oss << "Invalid port number in: [" << host << "]";
            errMsg = oss.str();
            return(false);
         }
      }

      if((port = atoi(theString.c_str())) == 0)
      {
         oss <<
            "Invalid port: [" << theString << "] " <<
            "in: [" << first << "]";

         errMsg = oss.str();
         return(false);
      }

      host = Util::substring(host, 0, index);
   }

   return(true);
}

static int clientThread(ThreadData *threadData)
{
   // We'll identify our output messages (if any) by
   // thread id so save that id off now

   int tID = GetCurrentThreadId();

   /*
    * We "own" the passed-in pointer 'threadData' and are
    * responsible for calling delete on it before we
    * depart from 'clientThread'.  Use an auto_ptr to ensure
    * 'threadData' is always deleted.  Note that this is
    * also how we ensure that our client socket will
    * always be closed.
    */

   auto_ptr<ThreadData> td(threadData);

   // Arrange for this thread's beginning/end to be noted

   ClientThreadHandler handler(
      td->number(),
      td->clientHandle(),
      td->mutex(),
      td->output(),
      tID
   );

   // Arrange to wait for some data to come in over our client connection
   // or until our timeout expires, whichever comes first

   ostringstream  oss;
   fd_set         rfds;
   struct timeval timeout;

   setTimeOut(td->timeout(), &timeout);

   FD_ZERO(&rfds);
   FD_SET((unsigned int)td->clientHandle(), &rfds);

   Output *o = td->output();

   oss <<
      "Calling select, timeout: [" << td->timeout() << "]";

   o->writeVerbose(_OFLI(tID, toString(oss)));

   if(select(td->clientHandle() + 1, &rfds, NULL, NULL, &timeout) <= 0)
   {
      oss <<
         "Call to 'select' timed out: [" << td->timeout() << "] seconds";

      o->write(_OFLI(tID, oss.str()));
      return(0);
   }

   ConnectLog connectLog(
      td->directory(), td->ip(), tID, td->number()
   );

   int n             = 0;
   char buffer[8192] = { '\0' };
   int bSize         = USZ(buffer);
   int bMax          = bSize - 1;

   // Read (what we believe to be) headers ...

   if((n = recv(td->clientHandle(), buffer, bMax, 0)) <= 0)
   {
      o->write(_OFLI(tID, "Initial call to 'recv' failed"));
      return(0);
   }

   oss << "Initial Read (headers) " << n << " bytes";
   o->writeVerbose(_OFLI(tID, toString(oss)));

   bool go         = true;
   string errMsg   = "";
   int returnValue = 0;

   while(go)
   {
      bool isConnect = false;
      string host    = "";
      string type    = "";
      int port       = 0;

      // Note we pass 'threadData' here instead of 'td' to avoid
      // any potential copy-constructor / auto_ptr / delete problems

      bool theStatus = processRequest(
         buffer,
         type,
         isConnect,
         host,
         port,
         o,
         tID,
         errMsg
      );

      if(!theStatus)
      {
         o->write(_OFLI(tID, errMsg));
         break;
      }

      oss << "called processRequest, host: " << host << ", port: " << port;
      o->writeVerbose(_OFLI(tID, toString(oss)));

      if(td->rlog())
      {
         if(!connectLog.beginSession(isConnect, host, port, errMsg) ||
            !connectLog.requestBytes(buffer, n, errMsg))
         {
            o->write(_OFLI(tID, errMsg));
            break;
         }
      }

      struct hostent *remoteHost = gethostbyname(host.c_str());

      if(!remoteHost)
      {
         oss << "Call to gethostbyname(" << host << ") failed";

         o->write(_OFLI(tID, oss.str()));
         break;
      }

      int remoteHandle = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);

      if(remoteHandle < 0)
      {
         oss << "Call to 'socket' failed, host: [" << host << "]";

         o->write(_OFLI(tID, oss.str()));
         break;
      }

      // Ensure 'remoteHandle' is cleaned up no matter what ...

      HandleCleanUp remoteCleanUp(remoteHandle);

      struct sockaddr_in remote_addr;

      remote_addr.sin_family = AF_INET;
      remote_addr.sin_port   = htons((unsigned short)port);

      memcpy(
         (void *)&remote_addr.sin_addr,
         (void *)remoteHost->h_addr,
         remoteHost->h_length
      );

      o->writeVerbose(_OFLI(tID, "Calling connect"));

      int status = connect(
         remoteHandle,
         (struct sockaddr *)&remote_addr,
         sizeof(remote_addr)
      );

      if(status < 0)
      {
         oss << "Could not connect to: [" << host << ":" << port << "]";

         o->write(_OFLI(tID, oss.str()));
         break;
      }

      o->writeVerbose(_OFLI(tID, "Back from connect"));

      int theHandle = remoteHandle;

      if(isConnect)
      {
         theHandle = td->clientHandle();
         strcpy(buffer, "HTTP/1.0 200 OK\r\n\r\n");
      }

      int theLength = strlen(buffer);

      if(send(theHandle, buffer, theLength, 0) != theLength)
      {
         oss << "(" << type << ") Error writing: " <<
            "[" << buffer << "], to host: [" << host << "], " <<
            "port: [" << port << "]";

         o->write(_OFLI(tID, oss.str()));
         break;
      }

      o->writeVerbose(_OFLI(tID, "Back from send, entering inner loop"));

      bool gotResponse = false;
      int ch           = td->clientHandle();

      // --------------------------------------------------------
      // Set 'go' to false so that by default, if we break
      // out of the while(true) loop below, we'll also break
      // out of while(go) above.  We'll reset 'go' to true
      // if we encounter a subsequent request. 
      // --------------------------------------------------------

      go = false;

      while(true)
      {
         FD_ZERO(&rfds);

         FD_SET((unsigned int)ch,           &rfds);
         FD_SET((unsigned int)remoteHandle, &rfds);
    
         int n = (ch > remoteHandle ? ch : remoteHandle);

         o->writeVerbose(_OFLI(tID, "Select: Begin"));

         int returnValue = select(n + 1, &rfds, NULL, NULL, &timeout);

         if(returnValue <= 0)
         {
            if(returnValue == 0)
            {
               oss << "Call to 'select' timed out (" <<
                  td->timeout() << ") seconds";
            }
            else
            {
               oss << "Call to 'select' failed";
            }

            o->write(_OFLI(tID, oss.str()));
            break;
         }

         o->writeVerbose(_OFLI(tID, "Select: End"));

         if(FD_ISSET(remoteHandle, &rfds))
         {
            o->writeVerbose(_OFLI(tID, "Remote Recv: Begin"));

            if((n = recv(remoteHandle, buffer, bMax, 0)) <= 0)
            {
               o->writeVerbose(_OFLI(tID, "Remote Recv: Exit"));

               returnValue = 1;
               break;
            }

            oss << "Remote Recv: End, " << n << " bytes";
            o->writeVerbose(_OFLI(tID, toString(oss)));

            if(td->rlog())
            {
               if(!connectLog.responseBytes(buffer, n, errMsg))
               {
                  o->write(_OFLI(tID, errMsg));
                  break;
               }
            }

            gotResponse = true;

            o->writeVerbose(_OFLI(tID, "Client Send: Begin"));

            if(send(td->clientHandle(), buffer, n, 0 ) != n)
            {
               oss << "Call to 'send' failed, host: [" << host <<
                  "], port: [" << port << "]";

               o->write(_OFLI(tID, oss.str()));
               break;
            }

            o->writeVerbose(_OFLI(tID, "Client Send: End"));
         }

         if(FD_ISSET(ch, &rfds))
         {
            o->writeVerbose(_OFLI(tID, "Client Recv: Begin"));

            if((n = recv(ch, buffer, bMax, 0)) <= 0)
            {
               o->writeVerbose(_OFLI(tID, "Client Recv: Exit"));

               returnValue = 1;
               break;
            }

            oss << "Client Recv: End, " << n << " bytes";
            o->writeVerbose(_OFLI(tID, toString(oss)));

            if(!isConnect && gotResponse)
            {
               /*
                * This is a new request so we head back
                * up to the 'while(go)' loop above where
                * we'll be handled by processRequest.
                * That function expects 'buffer' to be
                * null-terminated so ensure that's the case.
                */

               go        = true;
               buffer[n] = '\0';
               break;
            }

            if(td->rlog())
            {
               if(!connectLog.requestBytes(buffer, n, errMsg))
               {
                  o->write(_OFLI(tID, errMsg));
                  break;
               }
            }

            o->writeVerbose(_OFLI(tID, "Remote Send: Begin"));

            if(send(remoteHandle, buffer, n, 0 ) != n)
            {
               oss << "Call to 'send' failed, host: [" << host <<
                  "], port: [" << port << "]";

               o->write(_OFLI(tID, oss.str()));
               break;
            }

            o->writeVerbose(_OFLI(tID, "Remote Send: End"));
         }
      }
   }

   return(returnValue);
}

/*****************************************************************************
 * Main
 *****************************************************************************/

int main(int argc, char **argv)
{
   string errMsg;

   if(!goToWorkingDirectory(errMsg))
      bail(_OFL(errMsg));

   string fileName = (argc > 1 ? argv[1] : Config::PROGRAM_NAME + ".cfg");

   Config       config;
   MutexManager mgr;

   if(!getConfig(fileName, config, errMsg) || !mgr.setHandles(errMsg))
      bail(_OFL(errMsg));

   string outputDir = "";

   // If we're writing a log file or doing request/response logging
   // we need to determine our output directory and ensure
   // that it exists

   if(Cfg::isTrue(config.mLog) || Cfg::isTrue(config.mRLog))
   {
      if(!getOutputDirectory(outputDir, config.mDir, errMsg))
      {
         bail(_OFL(errMsg));
      }
   }

   Output *outputPtr = getOutput(
      mgr,
      outputDir,
      Cfg::isTrue(config.mVerbose),
      Cfg::isTrue(config.mConsole),
      Cfg::isTrue(config.mLog)
   );

   auto_ptr<Output> o(outputPtr);

   o->write(getBanner(config.mPort));

   o->writeVerbose(config.toString("Settings"));

   o->writeVerbose(_OFL("Output directory: " + outputDir));

   if(!doStartup(errMsg))
      o->exit(_OFL(errMsg));

   // Create our listening socket and bind to
   // the specified port number

   int proxyHandle = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);

   if(proxyHandle < 0)
      o->exit(_OFL("Call to 'socket' failed"));

   struct sockaddr_in proxy_addr;

   int portNumber             = atoi(config.mPort.c_str());
   proxy_addr.sin_family      = AF_INET;
   proxy_addr.sin_port        = htons((unsigned short)portNumber);
   proxy_addr.sin_addr.s_addr = INADDR_ANY;

   o->writeVerbose(_OFL("Calling bind"));

   int status = bind(
      proxyHandle,
      (struct sockaddr *)&proxy_addr,
      sizeof(proxy_addr)
   );

   if(status < 0)
      o->exit(_OFL("Call to 'bind' failed"));

   o->writeVerbose(_OFL("Bind successful, calling listen"));

   if(listen(proxyHandle, SOMAXCONN) != 0)
      o->exit(_OFL("Call to 'listen' failed"));

   o->writeVerbose(_OFL("Listen successful, entering main loop"));

   struct sockaddr_in        client_addr;
   int    size_client_addr = sizeof(client_addr);

   for(int number = 1; true; number ++)
   {
      o->writeVerbose(_OFL("Waiting for a connection"));
   
      int clientHandle = accept(
         proxyHandle,
         (struct sockaddr *)&client_addr,
         &size_client_addr
      );

      if(clientHandle < 0)
         o->exit(_OFL("Call to 'accept' failed"));

      unsigned int clientIP = client_addr.sin_addr.s_addr;

      ostringstream oss;

      oss <<
         ((clientIP >> 24) & 0xff) << "." <<
         ((clientIP >> 16) & 0xff) << "." <<
         ((clientIP >>  8) & 0xff) << "." <<
         ((clientIP >>  0) & 0xff);

      string ipString = oss.str();

      o->writeVerbose(
         _OFL("Connection accepted from: [" + ipString + "]")
      );

      ThreadData *tdPtr = new ThreadData(
         clientHandle,
         ipString,
         number,
         mgr.thread(),
         outputDir,
         Cfg::isTrue(config.mRLog),
         atof(config.mTimeout.c_str()),
         outputPtr
      );

      HANDLE theThread = CreateThread(
         NULL, 0, (LPTHREAD_START_ROUTINE)clientThread, tdPtr, 0, 0
      );

      if(!theThread)
         o->exit(_OFL("Call to 'CreateThread' failed"));

      CloseHandle(theThread);
   }

   return(0);
}
